import os
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.pyplot import cm, figure

from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.printing import stdout_success


class CircularPlotting(PlottingMixin):
    def __init__(self):
        PlottingMixin.__init__(self)

    def diffusion_plot(
        self,
        data: np.ndarray,
        fps: int,
        degree_width: int,
        palette: str,
        title: Optional[str] = None,
        save_path: Optional[Union[str, os.PathLike]] = None,
    ) -> plt.figure:
        """
        Create polar plot representing the angular diffusion within a video.

        .. image:: _static/img/circular_plotter.png
          :width: 450
          :align: center

        :param np.ndarray data: 1D np.ndarray with angle in degrees with one entry per frame.
        :param int fps: Framerate the video was recorded in.
        :param int degree_width: The width of the bars in the plot.
        :param str palette: The polar plot palette.
        :param str title: Title of the plot
        :param Optional[Union[str, os.PathLike]] save_path: Plot save location on disk. If None, then return plt.figure polar plot.

        :example:
        >>> data = np.random.normal(loc=180, scale=99, size=5000)
        >>> _ = CircularPlotting().diffusion_plot(data=data, title='Mean 180 degree plot', fps=30, degree_width=5, palette='jet', save_path='/Users/simon/Desktop/envs/troubleshooting/circular_features_zebrafish/project_folder/frames/output/dispersion/20200730_AB_7dpf_850nm_0004.png')
        """

        max_seconds = int(data.shape[0] / fps)
        second_bin = int(max_seconds / 5)
        if second_bin > 1:
            second_bin = 1
        data_rad = [x * 2 * np.pi / 360 for x in data]
        angle_bin_starts = np.arange(0.0, 2 * np.pi, 2 * np.pi * (degree_width / 360))
        n_length_bins = int(max_seconds / second_bin)
        bin_width = 2 * np.pi * (degree_width / 360)
        counts, bin_edges = np.histogram(data_rad, bins=angle_bin_starts)
        colors = self.create_single_color_lst(
            pallete_name=palette, increments=bin_edges.shape[0], as_rgb_ratio=True
        )
        norm_counts = counts / (fps * second_bin)
        bin_numbers = [np.round(norm_counts * n_length_bins / max_seconds, 0)]
        bin_lengths = [x / degree_width for x in bin_numbers]

        plt.figure().clear()
        plt.close()
        fig = figure(figsize=(8, 8))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], polar=True)
        bars = ax.bar(
            angle_bin_starts[:-1], bin_lengths[0], width=bin_width, bottom=0.0
        )
        ax.set_theta_zero_location("N")
        ax.set_theta_direction(-1)
        ax.set_yticklabels([])
        for cnt, (r, bar) in enumerate(zip(bin_lengths[0], bars)):
            bar.set_facecolor(colors[cnt])
        if title is not None:
            plt.title(title)
        if save_path is not None:
            plt.savefig(save_path)
            stdout_success(msg=f"Diffusion plot {save_path} created!")

        return ax

    def diffusion_time_bin_plot(
        self,
        data: np.ndarray,
        fps: int,
        time_bin: int,
        degree_width: int,
        palette: str,
        save_path: Union[str, os.PathLike],
    ):
        """
        Create polar plots representing angular diffusion within each N second time-bin of the video.

        .. image:: _static/img/cicular_time_bins.png
          :width: 600
          :align: center


        :param np.ndarray data: 1D np.ndarray with angle in degrees with one entry per frame.
        :param int fps: Framerate the video was recorded in.
        :param int time_bin: The length of each time bin (one plot will be created per time bin).
        :param int degree_width: The width of the bars in the plot.
        :param str palette: The polar plot palette.
        :param Optional[Union[str, os.PathLike]] save_path: Plot save location on disk. If None, then return plt.figure polar plot.

        :example:
        >>> data = np.random.normal(loc=180, scale=99, size=5000)
        >>> _ = CircularPlotting().diffusion_time_bin_plot(data=data, fps=30, degree_width=40, palette='jet', save_path='/Users/simon/Desktop/envs/troubleshooting/circular_features_zebrafish/project_folder/frames/output/dispertion_time_series/20200730_AB_7dpf_850nm_0004', time_bin=10)
        """
        time_bin_frame_size = time_bin * fps
        split_data = np.array_split(data, data.shape[0] / time_bin_frame_size)
        if not os.path.isdir(save_path):
            os.makedirs(save_path)
        for bin_cnt, time_bin_data in enumerate(split_data):
            time_plot = self.diffusion_plot(
                data=time_bin_data,
                fps=fps,
                degree_width=degree_width,
                palette=palette,
                title=f"Time bin {bin_cnt+1}",
            )
            fig_save_path = os.path.join(save_path, f"Time_bin_{bin_cnt+1}.png")
            time_plot.figure.savefig(fig_save_path)
            plt.figure().clear()
            plt.close()
            plt.clf()
            stdout_success(msg=f"Diffusion plot {save_path} created!")


#
#
#         #data = np.split(data, data.shape[0])
#
#
#
#
# #data = read_df('/Users/simon/Desktop/envs/troubleshooting/circular_features_zebrafish/project_folder/csv/circular_features/20200730_AB_7dpf_850nm_0004.csv', file_type='csv', usecols=[f'Angle_degrees_Animal_1']).values.flatten()
#
# data = np.random.normal(loc=90, scale=99, size=5000)
#
#
# _ = CircularPlotting().diffusion_plot(data=data, title='test', fps=30, degree_width=20, palette='jet', save_path='/Users/simon/Desktop/envs/troubleshooting/circular_features_zebrafish/project_folder/frames/output/dispersion/20200730_AB_7dpf_850nm_0004.png')
#
# #_ = CircularPlotting().diffusion_time_bin_plot(data=data, fps=30, degree_width=40, palette='jet', save_path='/Users/simon/Desktop/envs/troubleshooting/circular_features_zebrafish/project_folder/frames/output/dispertion_time_series/20200730_AB_7dpf_850nm_0004', time_bin=10)
#
#
# print(np.rad2deg(circstd(np.deg2rad(data))))
# print(np.rad2deg(circmean(np.deg2rad(data))))
#
#
# config_path='/Users/simon/Desktop/envs/troubleshooting/circular_features_zebrafish/project_folder/project_config.ini'
# save_path='/Users/simon/Desktop/envs/troubleshooting/circular_features_zebrafish/project_folder/frames/output/20200730_AB_7dpf_850nm_0002.png'
# DW = 20
# MAX_SECONDS = 10
# SECOND_BIN = 1
# LAST_BIN = 2*np.pi
#
# length_raw = np.arange(0,MAX_SECONDS+SECOND_BIN, SECOND_BIN)
# length_dict = {x: x/MAX_SECONDS for x in length_raw}
# data_rad = [x*2*np.pi/360 for x in data]
# angle_bin_starts = np.arange(0.0, 2*np.pi, 2*np.pi * (DW/360))
# n_length_bins = int(MAX_SECONDS/SECOND_BIN)
# bin_width = 2*np.pi * (DW/360)
#
# # for bin_start in angle_bin_starts:
# counts, bin_edges = np.histogram(data_rad, bins=angle_bin_starts)
# colors = PlottingMixin().create_single_color_lst(pallete_name='jet', increments=bin_edges.shape[0], as_rgb_ratio=True)
# norm_counts = counts/(30*SECOND_BIN)
# bin_numbers = [np.round(norm_counts * n_length_bins/MAX_SECONDS, 0)]
# bin_lengths = [x/5 for x in bin_numbers]
# bin_widths = [bin_width for idx in range(0, 72)]
# fig = figure(figsize=(8,8))
# ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], polar=True)
# #ax.set_yticklabels([list(range(1, MAX_SECONDS))])
# bars = ax.bar(angle_bin_starts[:-1], bin_lengths[0], width=bin_width, bottom=0.0)
# ax.set_theta_zero_location('N')
# ax.set_theta_direction(-1)
# ax.set_yticklabels([])
# #labels = [str(int(x)) + 's' for x in np.linspace(0, MAX_SECONDS, 4)]
# #labels_cnt = np.linspace(0, 1, 4)
#
# #ax.set_rgrids(labels_cnt, labels, fontsize=16)
# #lines, labels = ax.set_rgrids(labels_cnt, labels, fontsize=16)
# for cnt, (r, bar) in enumerate(zip(bin_lengths[0], bars)):
#     bar.set_facecolor(colors[cnt])
#     #bar.set_facecolor([cm.Reds(r)])
# #
# #
# #
#
# # #plt.clf()
# # compass_plot = plt.subplot(1, 1, 1, projection='polar')
# # compass_plot.set_theta_zero_location('N')
# # bars = compass_plot.bar(data, 0.01, width=0.01, bottom=0.0)
# #
# # # N = 20
# # theta = np.arange(0.0, 2*np.pi, 2*np.pi/N)
# # radii = 10*np.random.rand(N)
# # width = np.pi/4*np.random.rand(N)
# # bars = ax.bar(theta, radii, width=width, bottom=0.0)
# # for r,bar in zip(radii, bars):
# #     bar.set_facecolor( cm.jet(r/10.))
# #     bar.set_alpha(0.5)
# #
# # show()



import pytest
from simba.plotting.clf_validator import ClassifierValidationClips

class TestValidationClips(object):

    @pytest.fixture(params=['tests/data/test_projects/two_c57/project_folder/project_config.ini'])
    def config_path_args(self, request):
        return request

    @pytest.fixture(params=[['tests/data/test_projects/two_c57/project_folder/csv/machine_results/Together_1.csv']])
    def data_path_arg(self, request):
        return request

    @pytest.fixture(params=['Attack'])
    def clf_name_args(self, request):
        return request

    @pytest.fixture(params=[2])
    def window_args(self, request):
        return request

    @pytest.fixture(params=[True, False])
    def clip_args(self, request):
        return request

    @pytest.fixture(params=[True])
    def concat_args(self, request):
        return request

    @pytest.fixture(params=[(255, 0, 0)])
    def highlight_clr_args(self, request):
        return request

    @pytest.fixture(params=[0.5])
    def video_speed_args(self, request):
        return request

    @pytest.fixture(params=[(0, 0, 255)])
    def text_clr_args(self, request):
        return request


    def test_classifier_validation_clips(self,
                                         config_path_args,
                                         window_args,
                                         clf_name_args,
                                         data_path_arg,
                                         clip_args,
                                         concat_args,
                                         highlight_clr_args,
                                         video_speed_args,
                                         text_clr_args):

        clip_creator = ClassifierValidationClips(config_path=config_path_args.param,
                                                 window=window_args.param,
                                                 clf_name=clf_name_args.param,
                                                 data_paths=data_path_arg.param,
                                                 clips=clip_args.param,
                                                 concat_video=concat_args.param,
                                                 highlight_clr=highlight_clr_args.param,
                                                 video_speed=video_speed_args.param,
                                                 text_clr=text_clr_args.param)

        clip_creator.run()





# test = ClassifierValidationClips(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                  window=10,
#                                  clf_name='Attack',
#                                  clips=False,
#                                  concat_video=True,
#                                  highlight_clr=(255, 0, 0),
#                                  video_speed=0.5,
#                                  text_clr=(0, 0, 255))
# test.create_clips()



__author__ = "Simon Nilsson"

from copy import deepcopy
from typing import Dict, Optional

import numpy as np
import pandas as pd
from numba import jit, prange

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.roi_tools.ROI_analyzer import ROIAnalyzer
from simba.utils.read_write import get_fn_ext


class DirectingROIAnalyzer(ConfigReader, FeatureExtractionMixin):
    """
    Compute aggregate statistics for animals directing towards ROIs.

    :param str config_path: Path to SimBA project config file in Configparser format
    :param Optional[dict] settings: If dict, the animal body-parts and the probability threshold. If None, then the data is read from the
        project_config.ini. Defalt: None.

    .. note::
       `ROI tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.

    Examples
    ----------
    >>> settings = {'body_parts': {'Simon': 'Ear_left_1', 'JJ': 'Ear_left_2'}, 'threshold': 0.4}
    >>> _ = DirectingROIAnalyzer(config_path='MyProjectConfig', settings=settings).run()
    """

    def __init__(
        self,
        config_path: str,
        data_path: Optional[str] = None,
        settings: Optional[dict] = None,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        FeatureExtractionMixin.__init__(self, config_path=config_path)
        self.roi_analyzer = ROIAnalyzer(
            ini_path=config_path,
            data_path="outlier_corrected_movement_location",
            settings=settings,
        )
        self.files_found = deepcopy(self.roi_analyzer.files_found)
        self.direct_bp_dict = self.check_directionality_cords()

    def __format_direction_data(self):
        x_min = np.minimum(self.direction_data[:, 1], self.nose_arr[:, 0])
        y_min = np.minimum(self.direction_data[:, 2], self.nose_arr[:, 1])
        delta_x = abs((self.direction_data[:, 1] - self.nose_arr[:, 0]) / 2)
        delta_y = abs((self.direction_data[:, 2] - self.nose_arr[:, 1]) / 2)
        x_middle, y_middle = np.add(x_min, delta_x), np.add(y_min, delta_y)
        direction_data = np.concatenate(
            (y_middle.reshape(-1, 1), self.direction_data), axis=1
        )
        direction_data = np.concatenate(
            (x_middle.reshape(-1, 1), direction_data), axis=1
        )
        direction_data = np.delete(direction_data, [2, 3, 4], 1)
        bp_data = pd.DataFrame(
            direction_data, columns=["Eye_x", "Eye_y", "Directing_BOOL"]
        )
        bp_data["ROI_x"] = self.center_cord[0]
        bp_data["ROI_y"] = self.center_cord[1]
        bp_data = bp_data[["Eye_x", "Eye_y", "ROI_x", "ROI_y", "Directing_BOOL"]]
        bp_data.insert(loc=0, column="ROI", value=self.shape_info["Name"])
        bp_data.insert(loc=0, column="Animal", value=self.animal_name)
        bp_data.insert(loc=0, column="Video", value=self.video_name)
        bp_data = bp_data.reset_index().rename(columns={"index": "Frame"})
        bp_data = bp_data[bp_data["Directing_BOOL"] == 1].reset_index(drop=True)
        return bp_data

    @staticmethod
    @jit(nopython=True, fastmath=True)
    def ccw(roi_lines: np.array, eye_lines: np.array, shape_type: str):
        def calc(A, B, C):
            return (C[1] - A[1]) * (B[0] - A[0]) > (B[1] - A[1]) * (C[0] - A[0])

        results = np.full((eye_lines.shape[0], 4), -1)
        for i in prange(eye_lines.shape[0]):
            eye, roi = eye_lines[i][0:2], eye_lines[i][2:4]
            min_distance = np.inf

            if shape_type == "Circle":
                reversed_roi_lines = roi_lines[::-1]
                for j in prange(roi_lines.shape[0]):
                    dist_1 = np.sqrt(
                        (eye[0] - roi_lines[j][0]) ** 2
                        + (eye[1] - roi_lines[j][1]) ** 2
                    )
                    dist_2 = np.sqrt(
                        (eye[0] - roi_lines[j][2]) ** 2
                        + (eye[1] - roi_lines[j][3]) ** 2
                    )
                    if (dist_1 < min_distance) or (dist_2 < min_distance):
                        min_distance = min(dist_1, dist_2)
                        results[i] = reversed_roi_lines[j]

            else:
                for j in prange(roi_lines.shape[0]):
                    line_a, line_b = roi_lines[j][0:2], roi_lines[j][2:4]
                    center_x, center_y = (
                        line_a[0] + line_b[0] // 2,
                        line_a[1] + line_b[1] // 2,
                    )
                    if calc(eye, line_a, line_b) != calc(roi, line_a, line_b) or calc(
                        eye, roi, line_a
                    ) != calc(eye, roi, line_b):
                        distance = np.sqrt(
                            (eye[0] - center_x) ** 2 + (eye[1] - center_y) ** 2
                        )
                        if distance < min_distance:
                            results[i] = roi_lines[j]
                            min_distance = distance

        return results

    def __find_roi_intersections(self):
        eye_lines = self.bp_data[["Eye_x", "Eye_y", "ROI_x", "ROI_y"]].values.astype(
            int
        )
        roi_lines = None
        if self.shape_info["Shape_type"] == "Rectangle":
            top_left_x, top_left_y = (
                self.shape_info["topLeftX"],
                self.shape_info["topLeftY"],
            )
            bottom_right_x, bottom_right_y = (
                self.shape_info["Bottom_right_X"],
                self.shape_info["Bottom_right_Y"],
            )
            top_right_x, top_right_y = top_left_x + self.shape_info["width"], top_left_y
            bottom_left_x, bottom_left_y = (
                bottom_right_x - self.shape_info["width"],
                bottom_right_y,
            )
            roi_lines = np.array(
                [
                    [top_left_x, top_left_y, bottom_left_x, bottom_left_y],
                    [bottom_left_x, bottom_left_y, bottom_right_x, bottom_right_y],
                    [bottom_right_x, bottom_right_y, top_right_x, top_right_y],
                    [top_right_x, top_right_y, top_left_x, top_left_y],
                ]
            )

        elif self.shape_info["Shape_type"] == "Polygon":
            roi_lines = np.full((self.shape_info["vertices"].shape[0], 4), np.nan)
            roi_lines[-1] = np.hstack(
                (self.shape_info["vertices"][0], self.shape_info["vertices"][-1])
            )
            for i in range(self.shape_info["vertices"].shape[0] - 1):
                roi_lines[i] = np.hstack(
                    (self.shape_info["vertices"][i], self.shape_info["vertices"][i + 1])
                )

        elif self.shape_info["Shape_type"] == "Circle":
            center = self.shape_info[["centerX", "centerY"]].values.astype(int)
            roi_lines = np.full((2, 4), np.nan)
            roi_lines[0] = np.array(
                [
                    center[0],
                    center[1] - self.shape_info["radius"],
                    center[0],
                    center[1] + self.shape_info["radius"],
                ]
            )
            roi_lines[1] = np.array(
                [
                    center[0] - self.shape_info["radius"],
                    center[1],
                    center[0] + self.shape_info["radius"],
                    center[1],
                ]
            )

        return self.ccw(
            roi_lines=roi_lines,
            eye_lines=eye_lines,
            shape_type=self.shape_info["Shape_type"],
        )

    def run(self):
        """
        Method to calculate directing-towards ROI data
        """

        self.results_lst = []
        for file_cnt, file_path in enumerate(self.files_found):
            _, self.video_name, _ = get_fn_ext(file_path)
            self.roi_analyzer.files_found = [file_path]
            self.roi_analyzer.run()
            video_data_df = self.roi_analyzer.data_df
            for animal_name in self.roi_analyzer.multi_animal_id_list:
                self.animal_name = animal_name
                animal_bp_names = (
                    self.animal_bp_dict[animal_name]["X_bps"]
                    + self.animal_bp_dict[animal_name]["Y_bps"]
                )
                animal_data_df = video_data_df[animal_bp_names]
                animal_direct_bps = self.direct_bp_dict[animal_name]
                self.ear_left_arr = animal_data_df[
                    [
                        animal_direct_bps["Ear_left"]["X_bps"],
                        animal_direct_bps["Ear_left"]["Y_bps"],
                    ]
                ].to_numpy()
                self.ear_right_arr = animal_data_df[
                    [
                        animal_direct_bps["Ear_right"]["X_bps"],
                        animal_direct_bps["Ear_right"]["Y_bps"],
                    ]
                ].to_numpy()
                self.nose_arr = animal_data_df[
                    [
                        animal_direct_bps["Nose"]["X_bps"],
                        animal_direct_bps["Nose"]["Y_bps"],
                    ]
                ].to_numpy()
                for _, row in self.roi_analyzer.video_recs.iterrows():
                    self.shape_info = row
                    center_cord = (
                        (int(row["topLeftX"] + (row["width"] / 2))),
                        (int(row["topLeftY"] + (row["height"] / 2))),
                    )
                    self.center_cord = np.asarray(center_cord)
                    self.direction_data = self.jitted_line_crosses_to_static_targets(
                        left_ear_array=self.ear_left_arr,
                        right_ear_array=self.ear_right_arr,
                        nose_array=self.nose_arr,
                        target_array=self.center_cord,
                    )
                    self.bp_data = self.__format_direction_data()
                    eye_roi_intersections = pd.DataFrame(
                        self.__find_roi_intersections(),
                        columns=[
                            "ROI_edge_1_x",
                            "ROI_edge_1_y",
                            "ROI_edge_2_x",
                            "ROI_edge_2_y",
                        ],
                    )
                    self.bp_data = pd.concat(
                        [self.bp_data, eye_roi_intersections], axis=1
                    )
                    self.results_lst.append(self.bp_data)
                for _, row in self.roi_analyzer.video_circs.iterrows():
                    self.shape_info = row
                    self.center_cord = np.asarray((row["centerX"], row["centerY"]))
                    self.direction_data = self.jitted_line_crosses_to_static_targets(
                        left_ear_array=self.ear_left_arr,
                        right_ear_array=self.ear_right_arr,
                        nose_array=self.nose_arr,
                        target_array=self.center_cord,
                    )
                    self.bp_data = self.__format_direction_data()
                    eye_roi_intersections = pd.DataFrame(
                        self.__find_roi_intersections(),
                        columns=[
                            "ROI_edge_1_x",
                            "ROI_edge_1_y",
                            "ROI_edge_2_x",
                            "ROI_edge_2_y",
                        ],
                    )
                    self.bp_data = pd.concat(
                        [self.bp_data, eye_roi_intersections], axis=1
                    )
                    self.results_lst.append(self.bp_data)
                for _, row in self.roi_analyzer.video_polys.iterrows():
                    self.shape_info = row
                    self.center_cord = np.asarray((row["Center_X"], row["Center_Y"]))
                    self.direction_data = self.jitted_line_crosses_to_static_targets(
                        left_ear_array=self.ear_left_arr,
                        right_ear_array=self.ear_right_arr,
                        nose_array=self.nose_arr,
                        target_array=self.center_cord,
                    )
                    self.bp_data = self.__format_direction_data()
                    eye_roi_intersections = pd.DataFrame(
                        self.__find_roi_intersections(),
                        columns=[
                            "ROI_edge_1_x",
                            "ROI_edge_1_y",
                            "ROI_edge_2_x",
                            "ROI_edge_2_y",
                        ],
                    )
                    self.bp_data = pd.concat(
                        [self.bp_data, eye_roi_intersections], axis=1
                    )
                    self.results_lst.append(self.bp_data)
        self.results_df = pd.concat(self.results_lst, axis=0)


#
# test = DirectingROIAnalyzer(config_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/mouse_open_field/project_folder/project_config.ini')
# test.run()
#



__author__ = "Simon Nilsson"

import itertools
import json
import os
from collections import defaultdict

import h5py
import numpy as np
import pandas as pd

from simba.data_processors.interpolation_smoothing import Interpolate, Smooth
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pose_importer_mixin import PoseImporterMixin
from simba.utils.enums import Methods
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_all_videos_in_project,
                                    get_video_meta_data, write_df)


class SLEAPImporterSLP(ConfigReader, PoseImporterMixin):
    """
    Class for importing SLEAP pose-estimation data into a SimBA project.

    .. note::
      Importing SLEAP .SLP files into SimBA come at long runtimes. For fater runtimes, use
      :meth:`simba.pose_importers.sleap_h5_importer.SLEAPImporterH5` or :meth:`simba.pose_importers.sleap_csv_importer.SLEAPImporterCSV`

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str data_folder: Path to folder containing SLEAP data in `csv` format.
    :parameter List[str] id_lst: Animal names. This will be ignored in one animal projects and default to ``Animal_1``.
    :parameter str interpolation_settings: String defining the pose-estimation interpolation method. OPTIONS: 'None', 'Animal(s): Nearest',
        'Animal(s): Linear', 'Animal(s): Quadratic','Body-parts: Nearest', 'Body-parts: Linear',
        'Body-parts: Quadratic'.
    :parameter str smoothing_settings: Dictionary defining the pose estimation smoothing method. EXAMPLE: {'Method': 'Savitzky Golay',
        'Parameters': {'Time_window': '200'}}.

    Example
    ----------
    >>> slp_importer = SLEAPImporterSLP(project_path="MyConfigPath", data_folder=r'MySLPDataFolder', actor_IDs=['Mouse_1', 'Mouse_2'], interpolation_settings="Body-parts: Nearest", smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
    >>> slp_importer.run()

    References
    ----------
    .. [1] Pereira et al., SLEAP: A deep learning system for multi-animal pose tracking, `Nature Methods`,
           2022.

    """

    def __init__(
        self,
        project_path: str,
        data_folder: str,
        id_lst: list,
        interpolation_settings: str,
        smoothing_settings: dict,
    ):
        ConfigReader.__init__(self, config_path=project_path, read_video_info=False)
        PoseImporterMixin.__init__(self)
        self.interpolation_settings, self.smoothing_settings = (
            interpolation_settings,
            smoothing_settings,
        )
        self.data_folder, self.id_lst = data_folder, id_lst
        self.import_log_path = os.path.join(
            self.logs_path, f"data_import_log_{self.datetime}.csv"
        )
        self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir)
        self.input_data_paths = self.find_data_files(
            dir=self.data_folder, extensions=[".slp"]
        )
        self.data_and_videos_lk = self.link_video_paths_to_data_paths(
            data_paths=self.input_data_paths, video_paths=self.video_paths
        )
        if self.pose_setting is Methods.USER_DEFINED.value:
            self.__update_config_animal_cnt()
        if self.animal_cnt > 1:
            self.check_multi_animal_status()
            self.animal_bp_dict = self.create_body_part_dictionary(
                self.multi_animal_status,
                self.id_lst,
                self.animal_cnt,
                self.x_cols,
                self.y_cols,
                self.p_cols,
                self.clr_lst,
            )
            self.update_bp_headers_file()
        print(f"Importing {len(list(self.data_and_videos_lk.keys()))} file(s)...")

    def __h5_to_dict(self, name, obj):
        attr = list(obj.attrs.items())
        if name == "metadata":
            jsonList = attr[1][1]
            jsonList = jsonList.decode("utf-8")
            final_dictionary = json.loads(jsonList)
            final_dictionary = dict(final_dictionary)
            return final_dictionary

    def __check_that_all_animals_exist_in_frame(self):
        existing_animals = list(self.frame_dict.keys())
        missing_animals = [
            x for x in range(self.animal_cnt) if x not in existing_animals
        ]
        for missing_animal in missing_animals:
            self.frame_dict[missing_animal] = (
                [0] * ((len(self.analysis_dict["ordered_bps"]))) * 3
            )

    def __fill_missing_indexes(self):
        missing_indexes = list(
            set(list(range(0, self.video_info["frame_count"])))
            - set(list(self.data_df.index))
        )
        missing_df = pd.DataFrame(
            0, index=missing_indexes, columns=self.analysis_dict["xyp_headers"]
        )
        self.data_df = pd.concat([self.data_df, missing_df], axis=0)

    def __run_interpolation(self):
        print(
            f"Interpolating missing values in video {self.video_name} (Method: {self.interpolation_settings})..."
        )
        _ = Interpolate(
            input_path=self.save_path,
            config_path=self.config_path,
            method=self.interpolation_settings,
            initial_import_multi_index=True,
        )

    def __run_smoothing(self):
        print(
            f'Performing {self.smoothing_settings["Method"]} smoothing on video {self.video_name}...'
        )
        Smooth(
            config_path=self.config_path,
            input_path=self.save_path,
            time_window=int(self.smoothing_settings["Parameters"]["Time_window"]),
            smoothing_method=self.smoothing_settings,
        )

    def run(self):
        self.analysis_dict = defaultdict(list)
        self.save_paths_lst = []
        for file_cnt, (video_name, video_data) in enumerate(
            self.data_and_videos_lk.items()
        ):
            print(f"Analysing {video_name}...")
            video_timer = SimbaTimer(start=True)
            self.video_name = video_name
            in_h5 = h5py.File(video_data["DATA"], "r")
            self.sleap_dict = in_h5.visititems(self.__h5_to_dict)
            self.video_info = get_video_meta_data(video_path=video_data["VIDEO"])
            self.analysis_dict["bp_names"] = []
            self.analysis_dict["ordered_ids"] = []
            self.analysis_dict["ordered_bps"] = []
            self.analysis_dict["xy_headers"] = []
            self.analysis_dict["xyp_headers"] = []
            self.analysis_dict["animals_in_each_frame"] = []
            for bp in self.sleap_dict["nodes"]:
                self.analysis_dict["bp_names"].append(bp["name"])
            for orderVar in self.sleap_dict["skeletons"][0]["nodes"]:
                self.analysis_dict["ordered_ids"].append((orderVar["id"]))
            for index in self.analysis_dict["ordered_ids"]:
                self.analysis_dict["ordered_bps"].append(
                    self.analysis_dict["bp_names"][index]
                )

            with h5py.File(video_data["DATA"], "r") as file:
                self.analysis_dict["frames"] = file["frames"][:]
                self.analysis_dict["instances"] = file["instances"][:]
                self.analysis_dict["predicted_points"] = np.reshape(
                    file["pred_points"][:], (file["pred_points"][:].size, 1)
                )
            self.analysis_dict["no_frames"] = len(self.analysis_dict["frames"])

            for c in itertools.product(self.id_lst, self.analysis_dict["ordered_bps"]):
                x, y, p = (
                    str("{}_{}_x".format(c[0], c[1])),
                    str("{}_{}_y".format(c[0], c[1])),
                    (str("{}_{}_p".format(c[0], c[1]))),
                )
                self.analysis_dict["xy_headers"].extend((x, y))
                self.analysis_dict["xyp_headers"].extend((x, y, p))
            #
            self.data_df = pd.DataFrame(columns=self.analysis_dict["xyp_headers"])
            frames_lst = [l.tolist() for l in self.analysis_dict["frames"]]
            self.analysis_dict["animals_in_each_frame"] = [
                x[4] - x[3] for x in frames_lst
            ]
            self.__create_tracks()

            self.initialize_multi_animal_ui(
                animal_bp_dict=self.animal_bp_dict,
                video_info=self.video_info,
                data_df=self.data_df,
                video_path=video_data["VIDEO"],
            )
            if self.animal_cnt > 1:
                self.multianimal_identification()
            self.save_path = os.path.join(
                os.path.join(self.input_csv_dir, f"{self.video_name}.{self.file_type}")
            )
            write_df(
                df=self.out_df,
                file_type=self.file_type,
                save_path=self.save_path,
                multi_idx_header=True,
            )
            if self.interpolation_settings != "None":
                self.__run_interpolation()
            if self.smoothing_settings["Method"] != "None":
                self.__run_smoothing()
            video_timer.stop_timer()
            stdout_success(
                msg=f"Video {video_name} data imported...",
                elapsed_time=video_timer.elapsed_time_str,
            )
        self.timer.stop_timer()
        stdout_success(
            msg="All SLEAP SLP data files imported",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __create_tracks(self):
        start_frame = 0
        for frame_cnt, frame in enumerate(range(self.analysis_dict["no_frames"])):
            frame_idx = self.analysis_dict["frames"][frame_cnt][2]
            self.frame_dict = {}
            print(
                "Restructuring SLEAP frame: {}/{}, Video: {}".format(
                    str(frame_cnt),
                    str(self.analysis_dict["no_frames"]),
                    str(self.video_name),
                )
            )
            self.cnt_animals_frm = self.analysis_dict["animals_in_each_frame"][frame]
            if self.cnt_animals_frm == 0:
                self.frame_dict[0] = [0] * len(self.analysis_dict["xyp_headers"])
                end_frame = start_frame + (
                    len(self.analysis_dict["ordered_bps"]) * self.cnt_animals_frm
                )

            else:
                end_frame = start_frame + (
                    len(self.analysis_dict["ordered_bps"]) * self.cnt_animals_frm
                )
                start_animal, end_animal = 0, len(self.analysis_dict["ordered_bps"])
                frame_arr = self.analysis_dict["predicted_points"][
                    start_frame:end_frame
                ]
                for instance_counter, animal in enumerate(range(self.cnt_animals_frm)):
                    currRow = []
                    animal_arr = frame_arr[start_animal:end_animal]
                    track_id = self.analysis_dict["instances"][instance_counter][4]
                    for bp in animal_arr:
                        currRow.extend((bp[0][0], bp[0][1], bp[0][4]))
                    self.frame_dict[track_id] = currRow
                    start_animal += len(self.analysis_dict["ordered_bps"])
                    end_animal += len(self.analysis_dict["ordered_bps"])

            if self.animal_cnt > 1:
                self.__check_that_all_animals_exist_in_frame()
            frame_lst = [
                item for sublist in list(self.frame_dict.values()) for item in sublist
            ]
            start_frame = end_frame
            try:
                self.data_df.loc[frame_idx] = frame_lst
            except ValueError:
                break

        self.data_df.fillna(0, inplace=True)
        self.__fill_missing_indexes()
        self.data_df.sort_index(inplace=True)
        self.data_df.columns = self.bp_headers


# test = SLEAPImporterSLP(project_path="/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/data',
#                    id_lst=['Simon', 'Nastacia', 'JJ', 'Sam', 'Liana'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}}) #Savitzky Golay
# test.run()
# if test.animals_no > 1:
#     test.visualize_sleap()
# test.save_df()
# test.perform_interpolation()
# test.perform_smothing()
# print('All SLEAP imports complete.')


# test = ImportSLEAP(project_path="/Users/simon/Desktop/envs/troubleshooting/sleap_cody/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_cody/import_data',
#                    actor_IDs=['Animal_1', 'Animal_2'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}}) #Savitzky Golay
# test.initate_import_slp()
# if test.animals_no > 1:
#     test.visualize_sleap()
# test.save_df()
# test.perform_interpolation()
# test.perform_smothing()
# print('All SLEAP imports complete.')
